@@ -16,31 +16,39 @@
  */
 package org.apache.felix.webconsole.internal.configuration;
 
-
 import java.io.IOException;
 import java.io.PrintWriter;
+import java.io.StringWriter;
+import java.util.ArrayList;
+import java.util.List;
 import java.util.Locale;
+import java.util.Map;
 
 import javax.servlet.ServletException;
 import javax.servlet.http.HttpServletRequest;
 import javax.servlet.http.HttpServletResponse;
 
-import org.apache.felix.webconsole.DefaultVariableResolver;
+import org.apache.felix.utils.json.JSONWriter;
 import org.apache.felix.webconsole.SimpleWebConsolePlugin;
+import org.apache.felix.webconsole.WebConsoleConstants;
 import org.apache.felix.webconsole.WebConsoleUtil;
 import org.apache.felix.webconsole.internal.OsgiManagerPlugin;
-import org.json.JSONException;
-import org.json.JSONObject;
+import org.apache.felix.webconsole.internal.Util;
+import org.apache.felix.webconsole.internal.misc.ServletSupport;
+import org.apache.felix.webconsole.spi.ConfigurationHandler;
+import org.apache.felix.webconsole.spi.ValidationException;
+import org.osgi.framework.BundleContext;
 import org.osgi.framework.Constants;
 import org.osgi.framework.InvalidSyntaxException;
 import org.osgi.service.cm.Configuration;
+import org.osgi.util.tracker.ServiceTracker;
 
 
 /**
  * The <code>ConfigManager</code> class is the Web Console plugin to
  * manage configurations.
  */
-public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManagerPlugin
+public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManagerPlugin, ServletSupport
 {
 
     private static final long serialVersionUID = 5021174538498622428L;
@@ -52,11 +60,13 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
     static final String PID_FILTER = "pidFilter"; //$NON-NLS-1$
     static final String PID = "pid"; //$NON-NLS-1$
     static final String FACTORY_PID = "factoryPid"; //$NON-NLS-1$
-    static final String PLACEHOLDER_PID = "[Temporary PID replaced by real PID upon save]"; //$NON-NLS-1$
+    static final String REFERER = "referer"; //$NON-NLS-1$
+    static final String FACTORY_CREATE = "factoryCreate"; //$NON-NLS-1$
 
     static final String ACTION_CREATE = "create"; //$NON-NLS-1$
     static final String ACTION_DELETE = "delete"; //$NON-NLS-1$
     static final String ACTION_APPLY = "apply"; //$NON-NLS-1$
+    static final String ACTION_UPDATE = "update"; //$NON-NLS-1$
     static final String ACTION_UNBIND = "unbind"; //$NON-NLS-1$
     static final String PROPERTY_LIST = "propertylist"; //$NON-NLS-1$
     static final String LOCATION = "$location"; //$NON-NLS-1$
@@ -64,159 +74,142 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
     static final String CONFIGURATION_ADMIN_NAME = "org.osgi.service.cm.ConfigurationAdmin"; //$NON-NLS-1$
     static final String META_TYPE_NAME = "org.osgi.service.metatype.MetaTypeService"; //$NON-NLS-1$
 
+    public static final String UNBOUND_LOCATION = "??unbound:bundle/location"; //$NON-NLS-1$
+
     // templates
     private final String TEMPLATE;
 
+    // service tracker for SPI
+    private ServiceTracker<ConfigurationHandler, ConfigurationHandler> spiTracker;
+
     /** Default constructor */
-    public ConfigManager()
-    {
+    public ConfigManager() {
         super(LABEL, TITLE, CATEGORY_OSGI, CSS);
 
         // load templates
         TEMPLATE = readTemplateFile( "/templates/config.html" ); //$NON-NLS-1$
     }
 
-    static final boolean isAllowedPid(final String pid)
-    {
-        for(int i = 0; i < pid.length(); i++)
-        {
-            final char c = pid.charAt(i);
-            if ( c == '&' || c == '<' || c == '>' || c == '"' || c == '\'' )
-            {
-                return false;
-            }
-        }
-        return true;
+    @Override
+    public void activate(final BundleContext bundleContext) {
+        super.activate(bundleContext);
+        this.spiTracker = new ServiceTracker<>(bundleContext, ConfigurationHandler.class.getName(), null);
+        this.spiTracker.open(true);
     }
 
 
-    private static final Locale getLocale( HttpServletRequest request )
-    {
-        try
-        {
-            return request.getLocale();
-        }
-        catch ( Throwable t )
-        {
-            // expected in standard OSGi Servlet 2.1 environments
-            // fallback to using the default locale
-            return Locale.getDefault();
+    @Override
+    public void deactivate() {
+        if ( this.spiTracker != null ) {
+            this.spiTracker.close();
+            this.spiTracker = null;
         }
+        super.deactivate();
     }
 
-
     /**
      * @see javax.servlet.http.HttpServlet#doPost(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
      */
-    protected void doPost( HttpServletRequest request, HttpServletResponse response ) throws IOException
-    {
+    @Override
+    protected void doPost( HttpServletRequest request, HttpServletResponse response ) throws IOException {
+        // service unavailable if config admin is not available
+        final ConfigAdminSupport cas = getConfigurationAdminSupport();
+        if ( cas == null ) {
+            response.sendError(503);
+            return;
+        }
+
         // needed multiple times below
         String pid = request.getParameter( ConfigManager.PID );
-        if ( pid == null )
-        {
+        if ( pid == null ) {
             String info = request.getPathInfo();
             pid = WebConsoleUtil.urlDecode( info.substring( info.lastIndexOf( '/' ) + 1 ) );
         }
+        // ignore this request if the PID is invalid / not provided
+        if ( pid == null || pid.length() == 0 || !ConfigurationUtil.isAllowedPid(pid)) {
+            response.sendError(400);
+            return;
+        }
 
         // the filter to select part of the configurations
-        String pidFilter = request.getParameter( PID_FILTER );
-
-        final ConfigAdminSupport cas = getConfigurationAdminSupport();
-
-        // ignore this request if the PID and/or configuration admin is missing
-        if ( pid == null || pid.length() == 0 || cas == null )
-        {
-            // should log this here !!
+        final String pidFilter = request.getParameter( PID_FILTER );
+        if ( pidFilter != null && !ConfigurationUtil.isAllowedPid(pidFilter) ) {
+            response.sendError(400);
             return;
         }
 
-        // ignore this request, if the PID is invalid
-        if ( ! isAllowedPid(pid) )
-        {
-            response.sendError(500);
-            return;
-        }
-        if ( pidFilter != null && ! isAllowedPid(pidFilter) )
-        {
-            response.sendError(500);
+        if ( request.getParameter( ACTION_APPLY ) != null ) {
+            if ( request.getParameter( ConfigManager.ACTION_DELETE ) != null ) { //$NON-NLS-1$
+                try {
+                    cas.deleteConfiguration( pid );
+                    Util.sendJsonOk(response);
+                } catch ( final ValidationException ve) {
+                    response.sendError(400, ve.getMessage());
+                }
+            }
+            else
+            {
+                final String propertyList = request.getParameter( ConfigManager.PROPERTY_LIST ); //$NON-NLS-1$
+                if ( propertyList == null ) {
+                    response.sendError(400);
+                    return;
+                }
+
+                try {
+                    cas.applyConfiguration( request, pid, propertyList.split(","), ACTION_UPDATE.equals(request.getParameter(ACTION_APPLY)));
+                    String redirect = pid;
+                    if (pidFilter != null) {
+                        redirect = redirect.concat("?").concat(PID_FILTER).concat("=").concat(pidFilter);
+                    }
+
+                    WebConsoleUtil.sendRedirect(request, response, redirect);
+                } catch ( final ValidationException ve) {
+                    response.sendError(400, ve.getMessage());
+                }
+            }
+
             return;
         }
 
         // the configuration to operate on (to be created or "missing")
-        Configuration config = null;
+        final Configuration config;
 
         // should actually apply the configuration before redirecting
-        if ( request.getParameter( ACTION_CREATE ) != null ) //$NON-NLS-1$
-        {
-            config = cas.getPlaceholderConfiguration( pid ); // ca.createFactoryConfiguration( pid, null );
+        if ( request.getParameter( ACTION_CREATE ) != null ) {
+            config = ConfigurationUtil.getPlaceholderConfiguration( pid );
             pid = config.getPid();
-        }
-        else if ( request.getParameter( ACTION_APPLY ) != null ) //$NON-NLS-1$
-        {
-            String redirect = cas.applyConfiguration( request, pid );
-            if ( redirect != null )
-            {
-                if (pidFilter != null) {
-                    redirect += '?' + PID_FILTER + '=' + pidFilter;
-                }
-
-                WebConsoleUtil.sendRedirect(request, response, redirect);
-            }
-            else
-            {
-                response.setContentType( "application/json" ); //$NON-NLS-1$
-                response.setCharacterEncoding( "UTF-8" ); //$NON-NLS-1$
-                response.getWriter().print( "{ \"status\": true }" ); //$NON-NLS-1$
-            }
-
-            return;
-        }
-
-        if ( config == null )
-        {
-            config = cas.getConfiguration( pid );
+        } else {
+            config = cas.findConfiguration( pid );
         }
 
         // check for configuration unbinding
-        if ( request.getParameter( ACTION_UNBIND ) != null ) //$NON-NLS-1$
+        if ( request.getParameter( ACTION_UNBIND ) != null )
         {
-            if ( config != null && config.getBundleLocation() != null )
-            {
-                config.setBundleLocation( null );
+            if ( config != null && config.getBundleLocation() != null ) {
+                config.setBundleLocation( UNBOUND_LOCATION );
 
-                // workaround for Felix Config Admin 1.2.8 not clearing dynamic
-                // bundle location when clearing static bundle location. In
-                // this case we first set the static bundle location to the
-                // dynamic bundle location and then try to set both to null
-                if ( config.getBundleLocation() != null )
-                {
-                    config.setBundleLocation( "??invalid:bundle/location" ); //$NON-NLS-1$
-                    config.setBundleLocation( null );
-                }
             }
-            response.setContentType( "application/json" ); //$NON-NLS-1$
-            response.setCharacterEncoding( "UTF-8" ); //$NON-NLS-1$
-            response.getWriter().print( "{ \"status\": true }" ); //$NON-NLS-1$
+            Util.sendJsonOk(response);
             return;
         }
 
         // send the result
         response.setContentType( "application/json" ); //$NON-NLS-1$
         response.setCharacterEncoding( "UTF-8" ); //$NON-NLS-1$
-        final Locale loc = getLocale( request );
+        final Locale loc = Util.getLocale( request );
         final String locale = ( loc != null ) ? loc.toString() : null;
-        cas.printConfigurationJson( response.getWriter(), pid, config, pidFilter, locale );
+        cas.getJsonSupport().printConfigurationJson( response.getWriter(), pid, config, pidFilter, locale );
     }
 
 
     /**
      * @see org.apache.felix.webconsole.AbstractWebConsolePlugin#doGet(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
      */
+    @Override
     protected void doGet( HttpServletRequest request, HttpServletResponse response )
-    throws ServletException, IOException
-    {
+            throws ServletException, IOException {
         // check for "post" requests from previous versions
-        if ( "true".equals(request.getParameter("post")) ) {
+        if ( "true".equals(request.getParameter("post")) ) { //$NON-NLS-1$ //$NON-NLS-2$
             this.doPost(request, response);
             return;
         }
@@ -255,17 +248,17 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
             }
 
             // check both PID and PID filter
-            if ( pid != null && !isAllowedPid(pid) )
+            if ( pid != null && !ConfigurationUtil.isAllowedPid(pid) )
             {
-                response.sendError(500);
+                response.sendError(400);
             }
-            if ( pidFilter != null && !isAllowedPid(pidFilter) )
+            if ( pidFilter != null && !ConfigurationUtil.isAllowedPid(pidFilter) )
             {
-                response.sendError(500);
+                response.sendError(400);
             }
 
 
-            final Locale loc = getLocale( request );
+            final Locale loc = Util.getLocale( request );
             final String locale = ( loc != null ) ? loc.toString() : null;
 
             final PrintWriter pw = response.getWriter();
@@ -277,7 +270,7 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
                 final StringBuffer sb = new StringBuffer();
                 if ( pid != null && pidFilter != null)
                 {
-                    sb.append("(&");
+                    sb.append("(&"); //$NON-NLS-1$
                 }
                 if ( pid != null )
                 {
@@ -303,7 +296,7 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
                     // configuration
                     final Configuration[] configs = ca.listConfigurations( filter );
                     boolean printComma = false;
-                    for(int i=0; i<configs.length; i++)
+                    for(int i=0; configs != null && i<configs.length; i++)
                     {
                         final Configuration config = configs[i];
                         if ( config != null )
@@ -312,7 +305,7 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
                             {
                                 pw.print( ',' );
                             }
-                            ca.printConfigurationJson( pw, config.getPid(), config, null, locale );
+                            ca.getJsonSupport().printConfigurationJson( pw, config.getPid(), config, null, locale );
                             printComma = true;
                         }
                     }
@@ -339,6 +332,7 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
     /**
      * @see org.apache.felix.webconsole.AbstractWebConsolePlugin#renderContent(javax.servlet.http.HttpServletRequest, javax.servlet.http.HttpServletResponse)
      */
+    @Override
     protected void renderContent( HttpServletRequest request, HttpServletResponse response ) throws IOException
     {
 
@@ -378,46 +372,58 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
         }
 
         // check both PID and PID filter
-        if ( pid != null && !isAllowedPid(pid) )
+        if ( pid != null && !ConfigurationUtil.isAllowedPid(pid) )
         {
-            response.sendError(500);
+            response.sendError(400);
         }
-        if ( pidFilter != null && !isAllowedPid(pidFilter) )
+        if ( pidFilter != null && !ConfigurationUtil.isAllowedPid(pidFilter) )
         {
-            response.sendError(500);
+            response.sendError(400);
         }
 
-        final Locale loc = getLocale( request );
+        final Locale loc = Util.getLocale( request );
         final String locale = ( loc != null ) ? loc.toString() : null;
 
 
-        JSONObject json = new JSONObject();
-        try
+        StringWriter json = new StringWriter();
+        JSONWriter jw = new JSONWriter(json);
+        jw.object();
+        final ConfigAdminSupport cas = getConfigurationAdminSupport();
+        // check for osgi installer plugin
+        @SuppressWarnings("unchecked")
+        final Map<String, Object> labelMap = (Map<String, Object>) request.getAttribute(WebConsoleConstants.ATTR_LABEL_MAP);
+        jw.key("jsonsupport").value( labelMap.containsKey("osgi-installer-config-printer") ); //$NON-NLS-1$
+        final boolean hasMetatype = cas.getMetaTypeSupport() != null;
+        jw.key("status").value( cas != null ? Boolean.TRUE : Boolean.FALSE); //$NON-NLS-1$
+        jw.key("metatype").value( hasMetatype ? Boolean.TRUE : Boolean.FALSE); //$NON-NLS-1$
+        boolean hasConfigs = true;
+        if ( cas != null )
         {
-            final ConfigAdminSupport ca = getConfigurationAdminSupport();
-            json.put("status", ca != null ? Boolean.TRUE : Boolean.FALSE); //$NON-NLS-1$
-            if ( ca != null )
-            {
-                ca.listConfigurations( json, pidFilter, locale, loc );
-                ca.listFactoryConfigurations( json, pidFilter, locale );
-            }
+            hasConfigs = cas.getJsonSupport().listConfigurations( jw, pidFilter, locale, loc );
+            cas.getJsonSupport().listFactoryConfigurations( jw, pidFilter, locale );
         }
-        catch (JSONException e)
-        {
-            throw new IOException(e.toString());
+        if ( !hasConfigs && !hasMetatype && cas != null ) {
+            jw.key("noconfigs").value(true); //$NON-NLS-1$
+        } else {
+            jw.key("noconfigs").value(false); //$NON-NLS-1$
         }
 
+        jw.endObject();
+
         // if a configuration is addressed, display it immediately
-        if ( request.getParameter( ACTION_CREATE ) != null && pid != null ) //$NON-NLS-1$
-        {
-            pid = PLACEHOLDER_PID; // new PlaceholderConfiguration( pid ).getPid();
+        if ( request.getParameter( ACTION_CREATE ) != null && pid != null ) {
+            pid = ConfigurationUtil.getPlaceholderPid();
         }
 
 
         // prepare variables
-        DefaultVariableResolver vars = ( ( DefaultVariableResolver ) WebConsoleUtil.getVariableResolver( request ) );
+        final String referer = request.getParameter( REFERER );
+        final boolean factoryCreate = "true".equals( request.getParameter(FACTORY_CREATE) ); //$NON-NLS-1$
+        final Map<String, Object> vars = ( ( Map<String, Object> ) WebConsoleUtil.getVariableResolver( request ) );
         vars.put( "__data__", json.toString() ); //$NON-NLS-1$
-        vars.put( "selectedPid", pid != null ? pid : ""); //$NON-NLS-1$ //$NON-NLS-2$
+        vars.put( "selectedPid", pid != null ? pid : "" ); //$NON-NLS-1$ //$NON-NLS-2$
+        vars.put( "configurationReferer", referer != null ? referer : "" ); //$NON-NLS-1$ //$NON-NLS-2$
+        vars.put( "factoryCreate", Boolean.valueOf(factoryCreate) ); //$NON-NLS-1$
         vars.put( "param.apply", ACTION_APPLY ); //$NON-NLS-1$
         vars.put( "param.create", ACTION_CREATE ); //$NON-NLS-1$
         vars.put( "param.unbind", ACTION_UNBIND ); //$NON-NLS-1$
@@ -428,16 +434,24 @@ public class ConfigManager extends SimpleWebConsolePlugin implements OsgiManager
         response.getWriter().print(TEMPLATE);
     }
 
-    private ConfigAdminSupport getConfigurationAdminSupport()
-    {
+    private ConfigAdminSupport getConfigurationAdminSupport() {
         Object configurationAdmin = getService( CONFIGURATION_ADMIN_NAME );
-        if ( configurationAdmin != null )
-        {
-            return new ConfigAdminSupport( this, this.getBundleContext(), configurationAdmin );
+        if ( configurationAdmin != null ) {
+            final List<ConfigurationHandler> handlers = new ArrayList<>();
+            Object[] services = this.spiTracker.getServices();
+            if (services != null) {
+                for(final Object o : services) {
+                    handlers.add((ConfigurationHandler)o);
+                }
+            }
+            return new ConfigAdminSupport( this, configurationAdmin, handlers );
         }
         return null;
     }
 
-
+    @Override
+    public BundleContext getBundleContext() {
+        return super.getBundleContext();
+    }
 }
 
